from copy import copy
from typing import Any, Dict, List, Union, cast

PATH_TYPE = List[Union[str, int]]


def path_put(data: Any, path: Union[PATH_TYPE, str], value: Any) -> Any:
    """
    >>> path_put({}, "foo", "bar")
    {'foo': 'bar'}

    >>> path_put({}, "foo.bar.baz", "zoom")
    {'foo': {'bar': {'baz': 'zoom'}}}

    >>> path_put({}, "boo", [1, 2, 3, 4])
    {'boo': [1, 2, 3, 4]}

    >>> path_put({"aa": {"bb": "cc"}}, "aa.cc", "dd")
    {'aa': {'bb': 'cc', 'cc': 'dd'}}

    >>> path_put({"aa": {"bb": "cc"}}, "zz", {"foo": "bar"})
    {'aa': {'bb': 'cc'}, 'zz': {'foo': 'bar'}}

    >>> path_put({"aa": {"bb": "cc"}}, "aa.foo", "bar")
    {'aa': {'bb': 'cc', 'foo': 'bar'}}

    >>> path_put({}, "aa.0", "bar")
    {'aa': ['bar']}

    >>> path_put({}, "aa.3", "bar")
    {'aa': [None, None, None, 'bar']}

    >>> path_put({}, "aa.1.2", "bar")
    {'aa': [None, [None, None, 'bar']]}

    >>> path_put({}, "aa.1", {"foo": "bar"})
    {'aa': [None, {'foo': 'bar'}]}

    >>> path_put({}, "aa.1.foo", "bar")
    {'aa': [None, {'foo': 'bar'}]}

    >>> path_put({}, "aa.1.foo", {"bar": "baz"})
    {'aa': [None, {'foo': {'bar': 'baz'}}]}

    >>> path_put({"aa": [{"foo": "bar"}]}, "aa.0", {"moo": "shoo"})
    {'aa': [{'foo': 'bar', 'moo': 'shoo'}]}

    >>> path_put({"aa": [{"foo": "bar"}]}, "aa.0.baz.1", {"moo": "shoo"})
    {'aa': [{'foo': 'bar', 'baz': [None, {'moo': 'shoo'}]}]}

    >>> path_put({}, [], {"foo": "bar"})
    {'foo': 'bar'}

    >>> path_put({"aa": "bb"}, "", {"foo": "bar"})
    {'aa': 'bb', 'foo': 'bar'}

    >>> path_put({"aa": "bb", "subthing": {"s1": "v1"}}, "subthing", {"s2": "v2"})
    {'aa': 'bb', 'subthing': {'s1': 'v1', 's2': 'v2'}}

    >>> path_put({"things": [{"s1": "v1"}]}, "things.0", {"s2": "v2"})
    {'things': [{'s1': 'v1', 's2': 'v2'}]}

    >>> path_put({"things": [{"s1": "v1", "sub": {"ss1": "sv1"}}]}, "things.0.sub", {"ss2": "sv2"})
    {'things': [{'s1': 'v1', 'sub': {'ss1': 'sv1', 'ss2': 'sv2'}}]}

    >>> path_put({"thing": {"thing_id": "foo", "subs": [{"sub_id": 1, "sprop1": 11}]}},\
            "thing.subs", [{"sub_id": 1, "sprop2": 22}])
    {'thing': {'thing_id': 'foo', 'subs': [{'sub_id': 1, 'sprop1': 11, 'sprop2': 22}]}}

    """
    if data is None:
        return value

    if not path:
        path = []

    if isinstance(path, str):
        rendered_path = cast(PATH_TYPE, path.split("."))
    else:
        rendered_path = path

    if isinstance(data, dict):
        _path_put_dict(data, rendered_path, value)

    elif isinstance(data, list):
        _path_put_list(data, rendered_path, value)

    else:
        raise ValueError(
            f"path_put can only be used on dicts or lists, got {str(type(data))}"
        )

    return data


def _path_put_dict(data: Dict, path: PATH_TYPE, value: Any) -> None:
    if not path:
        if isinstance(value, dict):
            data.update(value)
        return

    path0 = path.pop(0)
    if not path:
        if path0 in data and isinstance(data[path0], list) and isinstance(value, list):
            for a, b in zip(data[path0], value):
                a.update(b)

        elif path0 in data and isinstance(value, dict):
            data[path0].update(value)
        else:
            data[path0] = value
        return

    if path0 not in data:
        if path and (isinstance(path[0], int) or path[0].isnumeric()):
            data[path0] = []
        else:
            data[path0] = {}

    path_put(data[path0], path, value)


def _path_put_list(data: List, path: PATH_TYPE, value: Any) -> None:
    original_path = copy(path)

    try:
        path0 = int(path.pop(0))
    except (ValueError, IndexError):
        raise KeyError(str(original_path))

    if len(data) < path0 + 1:
        for _i in range((path0 + 1) - len(data)):
            data.append(None)

    if len(path):
        # more depth
        if data[path0] is None:
            data[path0] = [] if isinstance(path[0], int) or path[0].isnumeric() else {}

        path_put(data[path0], path, value)

    elif isinstance(value, dict):
        if data[path0] is None:
            data[path0] = {}

        if not isinstance(data[path0], dict):
            raise ValueError(
                f"{str(original_path)} is not a dict but value asking to be merged is"
            )

        data[path0].update(value)

    else:
        data[path0] = value



from typing import Any, ClassVar, Dict, List, Optional

import pytest
from pydantic import BaseModel

from pydantic_enhanced_serializer import FieldsetConfig

from .utils import assert_expected_rendered_fieldset_data


def test_no_config_default_behaviour() -> None:
    class Level2Item(BaseModel):
        l2var1: str
        l2var2: int
        l2var3: str

    class Level2ListItem(BaseModel):
        l2lvar1: str
        l2lvar2: str

    class Level1Response(BaseModel):
        var1: str
        var2: str
        var3: int
        var4: int
        item: Level2Item
        items: List[Level2ListItem]

    response = Level1Response(
        var1="foo",
        var2="bar",
        var3=3,
        var4=4,
        item=Level2Item(
            l2var1="l21",
            l2var2=2,
            l2var3="123",
        ),
        items=[
            Level2ListItem(
                l2lvar1="l2l1",
                l2lvar2="l2l2",
            ),
            Level2ListItem(
                l2lvar1="l2l1B",
                l2lvar2="l2l2B",
            ),
            Level2ListItem(
                l2lvar1="l2l1C",
                l2lvar2="l2l2C",
            ),
        ],
    )

    assert_expected_rendered_fieldset_data(
        response,
        [],
        {
            "var1": "foo",
            "var2": "bar",
            "var3": 3,
            "var4": 4,
            "item": {
                "l2var1": "l21",
                "l2var2": 2,
                "l2var3": "123",
            },
            "items": [
                {
                    "l2lvar1": "l2l1",
                    "l2lvar2": "l2l2",
                },
                {
                    "l2lvar1": "l2l1B",
                    "l2lvar2": "l2l2B",
                },
                {
                    "l2lvar1": "l2l1C",
                    "l2lvar2": "l2l2C",
                },
            ],
        },
    )


# Zero config means always return ALL fields
@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {"field1": "one", "field2": "two", "field3": "three"}),
        (["field1"], {"field1": "one", "field2": "two", "field3": "three"}),
        (["field1", "field2"], {"field1": "one", "field2": "two", "field3": "three"}),
        (
            ["field1", "field2", "field3"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
        (["no_such_field"], {"field1": "one", "field2": "two", "field3": "three"}),
        (
            ["no_such_field", "field1"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
        (
            ["no_such_field", "field1", "field2"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
    ),
)
def test_single_level_by_field_name_no_config(
    fields: List[str], expected: dict
) -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


# Any config means all fields must be asked for or NONE are returned
@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {}),
        (["field1"], {"field1": "one"}),
        (["field1", "field2"], {"field1": "one", "field2": "two"}),
        (
            ["field1", "field2", "field3"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
        (["no_such_field"], {}),
        (["no_such_field", "field1"], {"field1": "one"}),
        (["no_such_field", "field1", "field2"], {"field1": "one", "field2": "two"}),
    ),
)
def test_single_level_by_field_name_any_config(
    fields: List[str], expected: dict
) -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


# Default fieldset always returned regardless of other args
@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {"field3": "three"}),
        (["field1"], {"field1": "one", "field3": "three"}),
        (["field1", "field2"], {"field1": "one", "field2": "two", "field3": "three"}),
        (
            ["field1", "field2", "field3"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
        (["no_such_field"], {"field3": "three"}),
        (["no_such_field", "field1"], {"field1": "one", "field3": "three"}),
        (
            ["no_such_field", "field1", "field2"],
            {"field1": "one", "field2": "two", "field3": "three"},
        ),
    ),
)
def test_single_level_by_field_subset_default(
    fields: List[str], expected: dict
) -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field3"]})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


def test_named_fieldset() -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"fset1": ["field3"], "fset2": ["field1", "field2"]}
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(response, ["fset1"], {"field3": "three"})
    assert_expected_rendered_fieldset_data(
        response, ["fset2"], {"field1": "one", "field2": "two"}
    )
    assert_expected_rendered_fieldset_data(
        response,
        ["fset1", "fset2"],
        {"field1": "one", "field2": "two", "field3": "three"},
    )


def test_named_fieldset_and_named_field() -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"fset1": ["field3"], "fset2": ["field1", "field2"]}
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(
        response, ["fset1", "field1"], {"field1": "one", "field3": "three"}
    )


@pytest.mark.parametrize(
    "fields,expected",
    (
        (
            [],
            {
                "field1": "one",
                "field2": "two",
                "field3": "three",
                "sub": {"subfield1": "sub1", "subfield2": "sub2"},
            },
        ),
        (
            ["field1"],
            {
                "field1": "one",
                "field2": "two",
                "field3": "three",
                "sub": {"subfield1": "sub1", "subfield2": "sub2"},
            },
        ),
        (
            ["field1", "sub"],
            {
                "field1": "one",
                "field2": "two",
                "field3": "three",
                "sub": {"subfield1": "sub1", "subfield2": "sub2"},
            },
        ),
        (
            ["field1", "sub.subfield1"],
            {
                "field1": "one",
                "field2": "two",
                "field3": "three",
                "sub": {"subfield1": "sub1", "subfield2": "sub2"},
            },
        ),
        (
            ["field1", "sub.subfield1", "sub.subfield2"],
            {
                "field1": "one",
                "field2": "two",
                "field3": "three",
                "sub": {"subfield1": "sub1", "subfield2": "sub2"},
            },
        ),
    ),
)
def test_nested_by_field_name_no_config_both(fields: List[str], expected: dict) -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: SubModel

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=SubModel(
            subfield1="sub1",
            subfield2="sub2",
        ),
    )
    assert_expected_rendered_fieldset_data(response, fields, expected)


@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {}),
        (["field1"], {"field1": "one"}),
        (["field1", "sub"], {"field1": "one", "sub": {}}),
        (["field1", "sub.subfield1"], {"field1": "one", "sub": {"subfield1": "sub1"}}),
        (
            ["field1", "sub.subfield1", "sub.subfield2"],
            {"field1": "one", "sub": {"subfield1": "sub1", "subfield2": "sub2"}},
        ),
    ),
)
def test_nested_by_field_name_any_config_both(
    fields: List[str], expected: dict
) -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={})

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: SubModel

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=SubModel(
            subfield1="sub1",
            subfield2="sub2",
        ),
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


# Default fieldset always returned regardless of other args
@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {"field3": "three"}),
        (["field1"], {"field1": "one", "field3": "three"}),
        (["sub"], {"field3": "three", "sub": {"subfield2": "sub2"}}),
        (
            ["sub.subfield1"],
            {"field3": "three", "sub": {"subfield1": "sub1", "subfield2": "sub2"}},
        ),
    ),
)
def test_nested_by_field_subset_default(fields: List[str], expected: dict) -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["subfield2"]})

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: SubModel

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field3"]})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=SubModel(
            subfield1="sub1",
            subfield2="sub2",
        ),
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


@pytest.mark.parametrize(
    "fields,expected",
    (
        ([], {"field3": "three", "sub": {"subfield2": "sub2"}}),
        (
            ["field1"],
            {"field1": "one", "field3": "three", "sub": {"subfield2": "sub2"}},
        ),
        (["sub"], {"field3": "three", "sub": {"subfield2": "sub2"}}),
        (
            ["sub.subfield1"],
            {"field3": "three", "sub": {"subfield1": "sub1", "subfield2": "sub2"}},
        ),
    ),
)
def test_nested_by_field_in_default(fields: List[str], expected: dict) -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["subfield2"]})

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: SubModel

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"default": ["field3", "sub"]}
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=SubModel(
            subfield1="sub1",
            subfield2="sub2",
        ),
    )

    assert_expected_rendered_fieldset_data(response, fields, expected)


def test_nested_named_fieldset_and_named_field() -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "subfset1": ["subfield2"],
                "subfset2": ["subfield1", "subfield2"],
            }
        )

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: SubModel

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field3"],
            }
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=SubModel(
            subfield1="sub1",
            subfield2="sub2",
        ),
    )

    assert_expected_rendered_fieldset_data(
        response, ["sub.subfset1"], {"field3": "three", "sub": {"subfield2": "sub2"}}
    )


def test_nested_named_fieldset_sublist() -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "subfset1": ["subfield2"],
                "subfset2": ["subfield1", "subfield2"],
            }
        )

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: List[SubModel]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field3"],
            }
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=[
            SubModel(
                subfield1="sub1",
                subfield2="sub2",
            ),
            SubModel(
                subfield1="sub3",
                subfield2="sub4",
            ),
        ],
    )

    assert_expected_rendered_fieldset_data(
        response,
        ["sub.subfset1"],
        {"field3": "three", "sub": [{"subfield2": "sub2"}, {"subfield2": "sub4"}]},
    )


def test_nested_model_with_default() -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["subfield2"],
            }
        )

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: List[SubModel]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field3"],
            }
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=[
            SubModel(
                subfield1="sub1",
                subfield2="sub2",
            ),
            SubModel(
                subfield1="sub3",
                subfield2="sub4",
            ),
        ],
    )

    assert_expected_rendered_fieldset_data(
        response,
        ["sub"],
        {"field3": "three", "sub": [{"subfield2": "sub2"}, {"subfield2": "sub4"}]},
    )


def test_nested_model_with_default_at_all_levels() -> None:
    class SubModel(BaseModel):
        subfield1: str
        subfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["subfield2"],
            }
        )

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str
        sub: List[SubModel]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field3", "sub"],
            }
        )

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
        sub=[
            SubModel(
                subfield1="sub1",
                subfield2="sub2",
            ),
            SubModel(
                subfield1="sub3",
                subfield2="sub4",
            ),
        ],
    )

    assert_expected_rendered_fieldset_data(
        response,
        [],
        {"field3": "three", "sub": [{"subfield2": "sub2"}, {"subfield2": "sub4"}]},
    )


def test_nested_optional_model_not_given() -> None:
    class QueryModel(BaseModel):
        what: str

    class ResponseModel(BaseModel):
        items: List[str]
        query: Optional[QueryModel] = None

    response = ResponseModel(items=[])

    assert_expected_rendered_fieldset_data(response, [], {"items": [], "query": None})


def test_default_start() -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["*"]})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(
        response, [], {"field1": "one", "field2": "two", "field3": "three"}
    )


def test_fieldsets_as_string() -> None:
    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field1"]})

    response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(
        response, "field2,field3", {"field1": "one", "field2": "two", "field3": "three"}
    )


def test_fieldset_that_references_itself() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field1"],
                "field2": ["field2"],
            }
        )

    class ResponseModel(BaseModel):
        things: List[Thing]

    response = ResponseModel(
        things=[
            Thing(
                field1="one",
                field2="two",
            )
        ]
    )

    assert_expected_rendered_fieldset_data(
        response,
        "things.other,things.field2,things.field5",
        {"things": [{"field1": "one", "field2": "two"}]},
    )


def test_fieldset_that_references_itself_but_does_not_exist_as_a_field() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field1"],
                "does_not_exist": ["does_not_exist", "field2"],
            }
        )

    class ResponseModel(BaseModel):
        things: List[Thing]

    response = ResponseModel(
        things=[
            Thing(
                field1="one",
                field2="two",
            )
        ]
    )

    assert_expected_rendered_fieldset_data(
        response,
        "things.does_not_exist",
        {"things": [{"field1": "one", "field2": "two"}]},
    )


def test_fieldsets_dict() -> None:
    class Thing(BaseModel):
        str1: str
        str2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["str1"],
            }
        )

    class Container(BaseModel):
        things: Dict[str, Thing]

    container = Container(
        things={
            "cat1": Thing(str1="c1t1str1val", str2="c1t1str2val"),
            "cat2": Thing(str1="c2t1str1val", str2="c2t1str2val"),
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": {"str1": "c1t1str1val"},
                "cat2": {"str1": "c2t1str1val"},
            }
        },
    )


def test_fieldsets_dict_non_model_value() -> None:
    class Container(BaseModel):
        things: Dict[str, str]

    container = Container(
        things={
            "cat1": "boo",
            "cat2": "lal",
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": "boo",
                "cat2": "lal",
            }
        },
    )


def test_fieldsets_dict_list() -> None:
    class Thing(BaseModel):
        str1: str
        str2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["str1"],
            }
        )

    class Container(BaseModel):
        things: Dict[str, List[Thing]]

    container = Container(
        things={
            "cat1": [
                Thing(str1="c1t1str1val", str2="c1t1str2val"),
                Thing(str1="c1t2str1val", str2="c1t2str2val"),
            ],
            "cat2": [
                Thing(str1="c2t1str1val", str2="c2t1str2val"),
            ],
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": [
                    {"str1": "c1t1str1val"},
                    {"str1": "c1t2str1val"},
                ],
                "cat2": [
                    {"str1": "c2t1str1val"},
                ],
            }
        },
    )


def test_fieldsets_dict_list_non_model_value() -> None:
    class Container(BaseModel):
        things: Dict[str, List[str]]

    container = Container(
        things={
            "cat1": ["boo", "bar", "baz"],
            "cat2": ["lal", "whatever"],
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": ["boo", "bar", "baz"],
                "cat2": ["lal", "whatever"],
            }
        },
    )


def test_fieldsets_dict_dict() -> None:
    class Thing(BaseModel):
        str1: str
        str2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["str1"],
            }
        )

    class Container(BaseModel):
        things: Dict[str, Dict[str, Thing]]

    container = Container(
        things={
            "cat1": {"nest1": Thing(str1="c1t1str1val", str2="c1t1str2val")},
            "cat2": {"nest2": Thing(str1="c2t1str1val", str2="c2t1str2val")},
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": {"nest1": {"str1": "c1t1str1val"}},
                "cat2": {"nest2": {"str1": "c2t1str1val"}},
            }
        },
    )


def test_fieldsets_dict_list_dict() -> None:
    class Thing(BaseModel):
        str1: str
        str2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["str1"],
            }
        )

    class Container(BaseModel):
        things: Dict[str, List[Dict[str, Thing]]]

    container = Container(
        things={
            "cat1": [
                {"nest1": Thing(str1="c1t1str1val", str2="c1t1str2val")},
                {"nest2": Thing(str1="c1t2str1val", str2="c1t2str2val")},
            ],
            "cat2": [
                {"nest3": Thing(str1="c3t1str1val", str2="c3t1str2val")},
            ],
        }
    )

    assert_expected_rendered_fieldset_data(
        container,
        "things.str1",
        {
            "things": {
                "cat1": [
                    {"nest1": {"str1": "c1t1str1val"}},
                    {"nest2": {"str1": "c1t2str1val"}},
                ],
                "cat2": [
                    {"nest3": {"str1": "c3t1str1val"}},
                ],
            }
        },
    )


def test_optional_dict_none() -> None:
    class Thing(BaseModel):
        str1: str
        data: Optional[Dict[str, Any]] = None

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["*"]})

    thing = Thing(str1="foo")

    assert_expected_rendered_fieldset_data(
        thing,
        "",
        {
            "str1": "foo",
            "data": None,
        },
    )



import asyncio
import inspect
from functools import wraps
from json.decoder import JSONDecodeError
from typing import Any, Callable, List, Optional

from fastapi import APIRouter as BaseAPIRouter
from fastapi import Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from pydantic import BaseModel, ValidationError

from ..render import render_fieldset_model


class APIRouter(BaseAPIRouter):
    def __init__(
        self,
        *args: Any,
        request_fields_name: str = "fields",
        maximum_expansion_depth: int = 5,
        raise_error_on_expansion_not_found: bool = False,
        **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.serializer_request_fields_name = request_fields_name
        self.serializer_maximum_expansion_depth = maximum_expansion_depth
        self.serializer_raise_error_on_expansion_not_found = (
            raise_error_on_expansion_not_found
        )

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        maximum_expansion_depth: Optional[int] = None,
        raise_error_on_expansion_not_found: Optional[bool] = None,
        **kwargs: Any
    ) -> None:
        # Wrap the endpoint function with actions to get fields from
        # the request and alter the render of the result based on it.

        # At runtime, we will need a request object to get the
        # fields parameter.  If the endpoint does not have a request
        # parameter, then add one in that we can use and hide it from
        # the real endpoint at runtime.
        signature = inspect.signature(endpoint)
        request_in_original_signature = False

        for param in signature.parameters.values():
            if issubclass(param.annotation, Request):
                request_param = param
                request_in_original_signature = True
                break

        if not request_in_original_signature:
            request_param = inspect.Parameter(
                name="_fsaaar_request",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Request,
            )

        @wraps(endpoint)
        async def field_aware_endpoint_wrapper(
            *endpoint_args: Any, **endpoint_kwargs: Any
        ) -> Any:
            request = endpoint_kwargs[request_param.name]
            if not request_in_original_signature:
                # Hide request object from real endpoint
                del endpoint_kwargs[request_param.name]

            result = endpoint(*endpoint_args, **endpoint_kwargs)
            if asyncio.iscoroutine(result):
                result = await result

            fields_request = await self._get_fields_from_request(request)

            if isinstance(result, BaseModel):
                content = await render_fieldset_model(
                    model=result,
                    fieldsets=fields_request,
                    maximum_expansion_depth=(
                        maximum_expansion_depth
                        if maximum_expansion_depth is not None
                        else self.serializer_maximum_expansion_depth
                    ),
                    raise_error_on_expansion_not_found=(
                        raise_error_on_expansion_not_found
                        if raise_error_on_expansion_not_found is not None
                        else self.serializer_raise_error_on_expansion_not_found
                    ),
                    expansion_context=request,
                    exclude_unset=kwargs.get("response_model_exclude_unset", False),
                    exclude_defaults=kwargs.get(
                        "response_model_exclude_defaults", False
                    ),
                    exclude_none=kwargs.get("response_model_exclude_none", False),
                )

                return JSONResponse(content=content)

            return result

        if not request_in_original_signature:
            # Add request to signature
            new_signature = signature.replace(
                parameters=[request_param, *signature.parameters.values()]
            )
            field_aware_endpoint_wrapper.__signature__ = new_signature  # type: ignore

        return super().add_api_route(path, field_aware_endpoint_wrapper, **kwargs)

    async def _get_fields_from_request(self, request: Request) -> List[str]:
        raw_fields: Any = None
        from_body: bool = False

        body_bytes = await request.body()
        if body_bytes:
            try:
                data = await request.json()
                raw_fields = data.get(self.serializer_request_fields_name)
                from_body = True
            except JSONDecodeError:
                pass

        if not raw_fields:
            raw_fields = [
                f
                for f in request.query_params.getlist(
                    self.serializer_request_fields_name
                )
                if f
            ]
            from_body = False

        if not raw_fields:
            return []

        if isinstance(raw_fields, str):
            return raw_fields.split(",")

        elif isinstance(raw_fields, list) and all(
            [isinstance(f, str) for f in raw_fields]
        ):
            fields = set()
            for field in raw_fields:
                fields.update(field.split(","))
            return list(fields)

        raise RequestValidationError(
            [
                ValidationError.from_exception_data(
                    title="Invalid value",
                    line_errors=[
                        {
                            "type": "invalid",
                            "msg": (  # type: ignore
                                "`{self.serializer_request_fields_name}` must be a "
                                "(optionally  comma saperated) string or "
                                "list of (optionally comma separated) strings"
                            ),
                            "loc": (
                                "body" if from_body else "query",
                                self.serializer_request_fields_name,
                            ),
                            "input": raw_fields,
                        }
                    ],
                )
            ]
        )



from asgiref.sync import async_to_sync
from django.http import HttpRequest
from ninja.parsers import Parser
from ninja.renderers import JSONRenderer
from pydantic import BaseModel

from pydantic_enhanced_serializer import render_fieldset_model


class PydanticFieldSetsRenderer(JSONRenderer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.fields_parameter_name = kwargs.get("fields_parameter_name", "fields")

    def pydantic_to_dict(
        self,
        data: BaseModel,
        request: HttpRequest,
        *,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> dict:
        # not very efficent to have to reparse the request...
        parser = Parser()
        fieldsets = (
            parser.parse_body(request).get(self.fields_parameter_name)
            or parser.parse_querydict(request.GET).get(self.fields_parameter_name)
            or []
        )

        result_data = async_to_sync(render_fieldset_model)(
            model=data,
            fieldsets=fieldsets,
            maximum_expansion_depth=5,
            raise_error_on_expansion_not_found=False,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return result_data



import datetime
from typing import Any, Awaitable, ClassVar, Dict, List, Optional

import pytest
from aiodataloader import DataLoader  # type: ignore
from pydantic import BaseModel, ConfigDict, Field

from pydantic_enhanced_serializer import FieldsetConfig, ModelExpansion

from .utils import assert_expected_rendered_fieldset_data


def test_singleton_expansion() -> None:
    class ExpandedModel(BaseModel):
        thing: str
        thing2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={})

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        def get_zoom(self, context: Any) -> ExpandedModel:
            return ExpandedModel(thing="what!", thing2="red")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(expansion_method_name="get_zoom"),
            }
        )

    api_response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(
        api_response,
        ["zoom.thing", "zoom.thing2"],
        {"zoom": {"thing": "what!", "thing2": "red"}},
    )


def test_nested_singleton_expansion() -> None:
    class SubExpandedModel(BaseModel):
        field1: str

    class ExpandedModel(BaseModel):
        thing: str
        thing2: str

        def get_sub(self, context: Any) -> SubExpandedModel:
            return SubExpandedModel(field1="foo")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "sub": ModelExpansion(
                    response_model=SubExpandedModel, expansion_method_name="get_sub"
                )
            },
        )

    class ResponseModel(BaseModel):
        field1: str
        field2: str
        field3: str

        def get_zoom(self, context: Any) -> ExpandedModel:
            return ExpandedModel(thing="what!", thing2="red")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    response_model=ExpandedModel, expansion_method_name="get_zoom"
                ),
            }
        )

    api_response = ResponseModel(
        field1="one",
        field2="two",
        field3="three",
    )

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom.sub"], {"zoom": {"sub": {"field1": "foo"}}}
    )


def test_dataloader_expansion() -> None:
    loader_call_count = 0

    class ExpandedModel(BaseModel):
        expanded_model_id: int
        field1: str

    async def batch_load_expanded_models(keys: List[int]):
        nonlocal loader_call_count
        loader_call_count += 1
        return list([ExpandedModel(expanded_model_id=k, field1=str(k)) for k in keys])

    class ItemDetail(BaseModel):
        item_id: int
        expanded_model_id: int
        fieldA: str

        def expand(self, context: dict) -> Awaitable:
            return context["dataloader"].load(self.item_id + 20)

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["item_id", "expanded_model_id", "fieldA"],
                "expanded_model": ModelExpansion(
                    expansion_method_name="expand",
                ),
            }
        )

    class ResponseModel(BaseModel):
        items: List[ItemDetail]

    api_response = ResponseModel(
        items=[
            ItemDetail(
                item_id=i,
                expanded_model_id=i + 20,
                fieldA=f"val{i}",
            )
            for i in range(5)
        ]
    )
    context = {"dataloader": DataLoader(batch_load_fn=batch_load_expanded_models)}

    assert_expected_rendered_fieldset_data(
        api_response,
        ["items.expanded_model"],
        {
            "items": [
                {
                    "item_id": i,
                    "expanded_model_id": i + 20,
                    "fieldA": f"val{i}",
                    "expanded_model": {
                        "expanded_model_id": i + 20,
                        "field1": str(i + 20),
                    },
                }
                for i in range(5)
            ]
        },
        context,
    )

    # Super important, make sure we are coalescing all the data loader calls, since
    # it is easy to break batching
    assert loader_call_count == 1


def test_multi_dataloader_expansion() -> None:
    loader1_call_count = 0
    loader2_call_count = 0

    class ExpandedModel(BaseModel):
        expanded_model_id: int
        field1: str

    class ExpandedModel2(BaseModel):
        expanded_model2_id: int
        field2: str

    async def batch_load_expanded_models(keys: List[int]) -> Any:
        nonlocal loader1_call_count
        loader1_call_count += 1

        return list([ExpandedModel(expanded_model_id=k, field1=str(k)) for k in keys])

    async def batch_load_expanded_models2(keys: List[int]) -> Any:
        nonlocal loader2_call_count
        loader2_call_count += 1

        return list([ExpandedModel2(expanded_model2_id=k, field2=str(k)) for k in keys])

    class ItemDetail(BaseModel):
        item_id: int
        expanded_model_id: int
        expanded_model2_id: int
        fieldA: str

        def expand_model(self, context: Any) -> Awaitable:
            return context["dataloader1"].load(self.item_id + 20)

        def expand_model2(self, context: Any) -> Awaitable:
            return context["dataloader2"].load(self.item_id + 40)

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["item_id", "expanded_model_id", "fieldA"],
                "expanded_model": ModelExpansion(
                    expansion_method_name="expand_model",
                ),
                "expanded_model2": ModelExpansion(
                    expansion_method_name="expand_model2",
                ),
            }
        )

    class ResponseModel(BaseModel):
        items: List[ItemDetail]

    api_response = ResponseModel(
        items=[
            ItemDetail(
                item_id=i,
                expanded_model_id=i + 20,
                expanded_model2_id=i + 40,
                fieldA=f"val{i}",
            )
            for i in range(5)
        ]
    )

    context = {
        "dataloader1": DataLoader(batch_load_fn=batch_load_expanded_models),
        "dataloader2": DataLoader(batch_load_fn=batch_load_expanded_models2),
    }

    assert_expected_rendered_fieldset_data(
        api_response,
        ["items.expanded_model", "items.expanded_model2"],
        {
            "items": [
                {
                    "item_id": i,
                    "expanded_model_id": i + 20,
                    "fieldA": f"val{i}",
                    "expanded_model": {
                        "expanded_model_id": i + 20,
                        "field1": str(i + 20),
                    },
                    "expanded_model2": {
                        "expanded_model2_id": i + 40,
                        "field2": str(i + 40),
                    },
                }
                for i in range(5)
            ]
        },
        context,
    )

    # Super important, make sure we are coalescing all the data loader calls, since
    # it is easy to break batching
    assert loader1_call_count == 1
    assert loader2_call_count == 1


def test_dataloader_expansion_nested() -> None:
    loader1_call_count = 0
    loader2_call_count = 0

    class ExpandedModel2(BaseModel):
        expanded_model2_id: int
        field2: str

    async def batch_load_expanded_models2(keys: List[int]) -> List[ExpandedModel2]:
        nonlocal loader2_call_count
        loader2_call_count += 1

        return list([ExpandedModel2(expanded_model2_id=k, field2=str(k)) for k in keys])

    class ExpandedModel(BaseModel):
        expanded_model_id: int
        expanded_model2_id: int
        field1: str

        def expand_model2(self, context: dict) -> Awaitable:
            return context["dataloader2"].load(self.expanded_model2_id)

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["expanded_model_id", "expanded_model2_id", "field1"],
                "expanded_model2": ModelExpansion(
                    expansion_method_name="expand_model2",
                ),
            }
        )

    async def batch_load_expanded_models(keys: List[int]) -> List[ExpandedModel]:
        nonlocal loader1_call_count
        loader1_call_count += 1

        return list(
            [
                ExpandedModel(
                    expanded_model_id=k, expanded_model2_id=k + 20, field1=str(k)
                )
                for k in keys
            ]
        )

    class ItemDetail(BaseModel):
        item_id: int
        expanded_model_id: int
        fieldA: str

        def expand_model(self, context: dict) -> Awaitable:
            return context["dataloader1"].load(self.expanded_model_id)

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["item_id", "expanded_model_id", "fieldA"],
                "expanded_model": ModelExpansion(
                    expansion_method_name="expand_model",
                ),
            }
        )

    class ResponseModel(BaseModel):
        items: List[ItemDetail]

    api_response = ResponseModel(
        items=[
            ItemDetail(
                item_id=i,
                expanded_model_id=i + 20,
                fieldA=f"val{i}",
            )
            for i in range(5)
        ]
    )

    context = {
        "dataloader1": DataLoader(batch_load_fn=batch_load_expanded_models),
        "dataloader2": DataLoader(batch_load_fn=batch_load_expanded_models2),
    }

    assert_expected_rendered_fieldset_data(
        api_response,
        ["items.expanded_model", "items.expanded_model.expanded_model2"],
        {
            "items": [
                {
                    "item_id": i,
                    "expanded_model_id": i + 20,
                    "fieldA": f"val{i}",
                    "expanded_model": {
                        "expanded_model_id": i + 20,
                        "expanded_model2_id": i + 40,
                        "field1": str(i + 20),
                        "expanded_model2": {
                            "expanded_model2_id": i + 40,
                            "field2": str(i + 40),
                        },
                    },
                }
                for i in range(5)
            ]
        },
        context,
    )

    # Super important, make sure we are coalescing all the data loader calls, since
    # it is easy to break batching
    assert loader1_call_count == 1
    assert loader2_call_count == 1


def test_merge_upwards_models() -> None:
    class ExpandedModel(BaseModel):
        thing: str
        thing2: str

    class ResponseModel(BaseModel):
        def get_zoom(self, context: Any) -> ExpandedModel:
            return ExpandedModel(thing="what!", thing2="red")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    expansion_method_name="get_zoom", merge_fields_upwards=True
                )
            }
        )

    api_response = ResponseModel()

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom.field1", "zoom"], {"thing": "what!", "thing2": "red"}
    )


def test_merge_upwards_dict() -> None:
    class ResponseModel(BaseModel):
        def get_zoom(self, context: Any) -> dict:
            return {"thing": "what!", "thing2": "red"}

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    expansion_method_name="get_zoom", merge_fields_upwards=True
                )
            }
        )

    api_response = ResponseModel()

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom"], {"thing": "what!", "thing2": "red"}
    )


def test_merge_upwards_nested_models() -> None:
    class SubExpandedModel(BaseModel):
        field1: str
        field2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field2"]})

    class ExpandedModel(BaseModel):
        thing: str
        thing2: str

        def get_sub(self, context: Any) -> SubExpandedModel:
            return SubExpandedModel(field1="f1", field2="f2")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "sub": ModelExpansion(
                    expansion_method_name="get_sub", merge_fields_upwards=True
                )
            }
        )

    class ResponseModel(BaseModel):
        def get_zoom(self, context: Any) -> ExpandedModel:
            return ExpandedModel(thing="what!", thing2="red")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    expansion_method_name="get_zoom", merge_fields_upwards=True
                )
            }
        )

    api_response = ResponseModel()

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom.sub.field1"], {"zoom": {"field1": "f1", "field2": "f2"}}
    )


def test_merge_upwards_nested_dicts() -> None:
    class ExpandedModel(BaseModel):
        thing: str
        thing2: str

        def get_sub(self, context: Any) -> dict:
            return {"field1": "f1", "field2": "f2"}

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "sub": ModelExpansion(
                    expansion_method_name="get_sub", merge_fields_upwards=True
                )
            }
        )

    class ResponseModel(BaseModel):
        def get_zoom(self, context: Any) -> ExpandedModel:
            return ExpandedModel(thing="what!", thing2="red")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    expansion_method_name="get_zoom", merge_fields_upwards=True
                )
            }
        )

    api_response = ResponseModel()

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom.sub.field1"], {"zoom": {"field1": "f1", "field2": "f2"}}
    )


def test_expand_to_scalar_value() -> None:
    class ResponseModel(BaseModel):
        def get_zoom(self, context: Any) -> str:
            return "some scalar value"

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"zoom": ModelExpansion(expansion_method_name="get_zoom")}
        )

    api_response = ResponseModel()

    assert_expected_rendered_fieldset_data(
        api_response, ["zoom"], {"zoom": "some scalar value"}
    )


def test_expand_to_scalar_value_with_merge() -> None:
    class SubModel(BaseModel):
        def get_zoom(self, context: Any) -> str:
            return "some scalar value"

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "zoom": ModelExpansion(
                    expansion_method_name="get_zoom", merge_fields_upwards=True
                )
            }
        )

    class ResponseModel(BaseModel):
        sub: SubModel

    api_response = ResponseModel(sub=SubModel())

    with pytest.raises(ValueError) as exc:
        assert_expected_rendered_fieldset_data(api_response, ["sub.zoom"], {})

    assert "merge_fields_upwards=True" in str(exc.value)


def test_merge_upwards_lists() -> None:
    class SubExpandedModel(BaseModel):
        field1: str
        field2: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field2"]})

    class SubModel(BaseModel):
        thing: str
        thing2: str

        def get_sub(self, context: Any) -> SubExpandedModel:
            return SubExpandedModel(field1=self.thing + "f1", field2=self.thing2 + "f2")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["thing", "thing2"],
                "sub": ModelExpansion(
                    expansion_method_name="get_sub", merge_fields_upwards=True
                ),
            }
        )

    class ResponseModel(BaseModel):
        items: List[SubModel]

    api_response = ResponseModel(
        items=[
            SubModel(thing="t1", thing2="t2"),
            SubModel(thing="t3", thing2="t4"),
        ]
    )

    assert_expected_rendered_fieldset_data(
        api_response,
        ["items.sub"],
        {
            "items": [
                {
                    "thing": "t1",
                    "thing2": "t2",
                    "field2": "t2f2",
                },
                {
                    "thing": "t3",
                    "thing2": "t4",
                    "field2": "t4f2",
                },
            ]
        },
    )


def test_expansion_returns_list() -> None:
    class ExpandedModel(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field1"]})

    class ResponseModel(BaseModel):
        f1: str

        def get_sub(self, context: Any) -> List[ExpandedModel]:
            return [
                ExpandedModel(field1="field1value1"),
                ExpandedModel(field1="field1value2"),
                ExpandedModel(field1="field1value3"),
            ]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["f1"],
                "sub": ModelExpansion(expansion_method_name="get_sub"),
            }
        )

    api_response = ResponseModel(f1="f1value")

    assert_expected_rendered_fieldset_data(
        api_response,
        ["sub"],
        {
            "f1": "f1value",
            "sub": [
                {"field1": "field1value1"},
                {"field1": "field1value2"},
                {"field1": "field1value3"},
            ],
        },
    )


def test_expansion_returns_list_nested() -> None:
    class SubExpandedModel(BaseModel):
        subfield1: str

    class ExpandedModel(BaseModel):
        field1: str

        def get_subsub(self, context: Any) -> List[SubExpandedModel]:
            return [
                SubExpandedModel(subfield1=f"{self.field1}_subvalue1"),
                SubExpandedModel(subfield1=f"{self.field1}_subvalue2"),
                SubExpandedModel(subfield1=f"{self.field1}_subvalue3"),
            ]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field1"],
                "subsub": ModelExpansion(expansion_method_name="get_subsub"),
            }
        )

    class ResponseModel(BaseModel):
        f1: str

        def get_sub(self, context: Any) -> List[ExpandedModel]:
            return [
                ExpandedModel(field1="field1value1"),
                ExpandedModel(field1="field1value2"),
                ExpandedModel(field1="field1value3"),
            ]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["f1"],
                "sub": ModelExpansion(expansion_method_name="get_sub"),
            }
        )

    api_response = ResponseModel(f1="f1value")

    assert_expected_rendered_fieldset_data(
        api_response,
        ["sub", "sub.subsub"],
        {
            "f1": "f1value",
            "sub": [
                {
                    "field1": "field1value1",
                    "subsub": [
                        {"subfield1": "field1value1_subvalue1"},
                        {"subfield1": "field1value1_subvalue2"},
                        {"subfield1": "field1value1_subvalue3"},
                    ],
                },
                {
                    "field1": "field1value2",
                    "subsub": [
                        {"subfield1": "field1value2_subvalue1"},
                        {"subfield1": "field1value2_subvalue2"},
                        {"subfield1": "field1value2_subvalue3"},
                    ],
                },
                {
                    "field1": "field1value3",
                    "subsub": [
                        {"subfield1": "field1value3_subvalue1"},
                        {"subfield1": "field1value3_subvalue2"},
                        {"subfield1": "field1value3_subvalue3"},
                    ],
                },
            ],
        },
    )


def test_expansion_returns_list_of_list() -> None:
    class ExpandedModel(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(fieldsets={"default": ["field1"]})

    class ResponseModel(BaseModel):
        f1: str

        def get_sub(self, context: Any) -> List[List[ExpandedModel]]:
            return [
                [
                    ExpandedModel(field1="field1value1"),
                    ExpandedModel(field1="field1value2"),
                ],
                [ExpandedModel(field1="field1value3")],
            ]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["f1"],
                "sub": ModelExpansion(expansion_method_name="get_sub"),
            }
        )

    api_response = ResponseModel(f1="f1value")

    assert_expected_rendered_fieldset_data(
        api_response,
        ["sub"],
        {
            "f1": "f1value",
            "sub": [
                [{"field1": "field1value1"}, {"field1": "field1value2"}],
                [{"field1": "field1value3"}],
            ],
        },
    )


def test_expand_empty_list() -> None:
    class ResponseModel(BaseModel):
        f1: str

        def get_sub(self, context: Any) -> List[int]:
            return []

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["f1"],
                "sub": ModelExpansion(expansion_method_name="get_sub"),
            }
        )

    api_response = ResponseModel(f1="f1value")

    assert_expected_rendered_fieldset_data(
        api_response,
        ["sub"],
        {"f1": "f1value", "sub": []},
    )


def test_expansion_in_default_fieldset() -> None:
    class SubModelBase(BaseModel):
        f1: str

    class SubModel(SubModelBase):
        f2: str
        f3: str
        f4: str
        f5: str
        f6: str
        f7: str

        def get_sub(self, context: Any) -> str:
            return "sub" + self.f1

        model_config = ConfigDict(from_attributes=True)

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["f1", "f2", "sub"],
                "g1": ["f3"],
                "g2": ["f4", "f5"],
                "sub": ModelExpansion(expansion_method_name="get_sub"),
            }
        )

    class ResponseModel(BaseModel):
        subs: List[SubModel]

    response = ResponseModel(
        subs=[
            SubModel(
                f1="f1.1",
                f2="f1.2",
                f3="f1.3",
                f4="f1.4",
                f5="f1.5",
                f6="f1.6",
                f7="f1.7",
            ),
            SubModel(
                f1="f2.1",
                f2="f2.2",
                f3="f2.3",
                f4="f2.4",
                f5="f2.5",
                f6="f2.6",
                f7="f2.7",
            ),
        ]
    )

    assert_expected_rendered_fieldset_data(
        response,
        ["subs.g1", "subs.g2", "subs.sub", "subs.f6", "subs.f7"],
        {
            "subs": [
                {
                    "f1": "f1.1",
                    "f2": "f1.2",
                    "f3": "f1.3",
                    "f4": "f1.4",
                    "f5": "f1.5",
                    "f6": "f1.6",
                    "f7": "f1.7",
                    "sub": "subf1.1",
                },
                {
                    "f1": "f2.1",
                    "f2": "f2.2",
                    "f3": "f2.3",
                    "f4": "f2.4",
                    "f5": "f2.5",
                    "f6": "f2.6",
                    "f7": "f2.7",
                    "sub": "subf2.1",
                },
            ]
        },
    )


def test_nested_array_expansion_overlap() -> None:
    # Expansions should not overwrite request fields where the request
    # field is a standalone field not part of other fieldset groups

    class SubEntity(BaseModel):
        addr: str

    class Struct(BaseModel):
        num_stuff: int

    class Entity(BaseModel):
        entity_id: str
        structure: Optional[Struct]
        created: datetime.datetime

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["entity_id"],
                "sub_entity": ModelExpansion(
                    response_model=SubEntity, expansion_method_name="get_sub_entity"
                ),
                "timestamps": ["created"],
            }
        )

        def get_sub_entity(self, context: Any) -> SubEntity:
            return SubEntity(addr="somewhere")

    now = datetime.datetime.now()

    class Thing(BaseModel):
        thing_id: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["thing_id"],
                "entities": ModelExpansion(
                    response_model=List[Entity],
                    expansion_method_name="get_entities",
                ),
            }
        )

        def get_entities(self, context: Any) -> List[Entity]:
            return [
                Entity(entity_id="foo", structure=Struct(num_stuff=11), created=now)
            ]

    class Response(BaseModel):
        thing: Thing

    response = Response(thing=Thing(thing_id="bar"))

    assert_expected_rendered_fieldset_data(
        response,
        [
            "thing.entities.sub_entity",
            "thing.entities.structure",
            "thing.entities.timestamps",
        ],
        {
            "thing": {
                "thing_id": "bar",
                "entities": [
                    {
                        "entity_id": "foo",
                        "sub_entity": {
                            "addr": "somewhere",
                        },
                        "structure": {
                            "num_stuff": 11,
                        },
                        "created": now,
                    }
                ],
            }
        },
    )


def test_nested_expansion_dict_vary_keys() -> None:
    class Storey(BaseModel):
        s_attr1: str
        component_counts: Optional[Dict[str, int]]

    class ItemStructure(BaseModel):
        attr1: str
        storeys: List[Storey] = Field(min_length=1)

    class Item(BaseModel):
        item_id: str
        structure: Optional[ItemStructure] = None

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["item_id"],
            }
        )

    class Thing(BaseModel):
        thing_id: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["thing_id"],
                "items": ModelExpansion(
                    response_model=List[Item],
                    expansion_method_name="get_items",
                ),
            }
        )

        def get_items(self, context: Any) -> List[Item]:
            return [
                Item(
                    item_id="1234",
                    structure=ItemStructure(
                        attr1="a_val1",
                        storeys=[
                            Storey(
                                s_attr1="Floor 1-1",
                                component_counts={
                                    "key11": 11,
                                    "key12": 12,
                                },
                            ),
                            Storey(
                                s_attr1="Floor 1-2",
                                component_counts={
                                    "key21": 21,
                                    "key22": 22,
                                    "key23": 23,
                                },
                            ),
                        ],
                    ),
                ),
                Item(
                    item_id="5678",
                    structure=ItemStructure(
                        attr1="a_val2",
                        storeys=[
                            Storey(
                                s_attr1="Floor 2-1",
                                component_counts={
                                    "key211": 211,
                                    "key212": 212,
                                },
                            ),
                            Storey(
                                s_attr1="Floor 2-2",
                                component_counts={
                                    "key221": 221,
                                    "key222": 222,
                                    "key223": 223,
                                },
                            ),
                            Storey(
                                s_attr1="Floor 3-2",
                                component_counts={
                                    "key321": 321,
                                    "key322": 322,
                                    "key323": 323,
                                    "key324": 324,
                                },
                            ),
                        ],
                    ),
                ),
            ]

    class Response(BaseModel):
        thing: Thing

    response = Response(
        thing=Thing(
            thing_id="abc",
        )
    )

    assert_expected_rendered_fieldset_data(
        response,
        [
            "thing.items.structure",
        ],
        {
            "thing": {
                "thing_id": "abc",
                "items": [
                    {
                        "item_id": "1234",
                        "structure": {
                            "attr1": "a_val1",
                            "storeys": [
                                {
                                    "s_attr1": "Floor 1-1",
                                    "component_counts": {
                                        "key11": 11,
                                        "key12": 12,
                                    },
                                },
                                {
                                    "s_attr1": "Floor 1-2",
                                    "component_counts": {
                                        "key21": 21,
                                        "key22": 22,
                                        "key23": 23,
                                    },
                                },
                            ],
                        },
                    },
                    {
                        "item_id": "5678",
                        "structure": {
                            "attr1": "a_val2",
                            "storeys": [
                                {
                                    "s_attr1": "Floor 2-1",
                                    "component_counts": {
                                        "key211": 211,
                                        "key212": 212,
                                    },
                                },
                                {
                                    "s_attr1": "Floor 2-2",
                                    "component_counts": {
                                        "key221": 221,
                                        "key222": 222,
                                        "key223": 223,
                                    },
                                },
                                {
                                    "s_attr1": "Floor 3-2",
                                    "component_counts": {
                                        "key321": 321,
                                        "key322": 322,
                                        "key323": 323,
                                        "key324": 324,
                                    },
                                },
                            ],
                        },
                    },
                ],
            }
        },
    )


def test_dict_nested_list_dicts() -> None:
    class Thing(BaseModel):
        str1: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["str1"],
                "subdata": ModelExpansion(
                    response_model=Optional[Dict[str, Any]],
                    expansion_method_name="get_subdata",
                ),
            }
        )

        def get_subdata(self, *args: Any, **kwargs: Any) -> Optional[Dict[str, Any]]:
            return {
                "list": [
                    {
                        "a": "b",
                    }
                ]
            }

    class ThingContainer(BaseModel):
        things: List[Thing]

    thing = Thing(
        str1="foo",
    )

    things = ThingContainer(things=[thing, thing])

    assert_expected_rendered_fieldset_data(
        things,
        "things.subdata",
        {
            "things": [
                {
                    "str1": "foo",
                    "subdata": {
                        "list": [
                            {
                                "a": "b",
                            }
                        ]
                    },
                },
                {
                    "str1": "foo",
                    "subdata": {
                        "list": [
                            {
                                "a": "b",
                            }
                        ]
                    },
                },
            ]
        },
    )



from asyncio import gather
from copy import copy
from typing import Any, List, Set, Union

from pydantic import BaseModel

from .fieldsets import fieldset_to_includes
from .models import ExpansionInstruction
from .path_put import path_put


async def render_fieldset_model(
    model: BaseModel,
    fieldsets: Union[str, Set[str], List[str]],
    maximum_expansion_depth: int = 5,
    raise_error_on_expansion_not_found: bool = False,
    expansion_context: Any = None,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
) -> dict:
    includes, expansions = fieldset_to_includes(fieldsets, model)
    rendered_model = model.model_dump(
        include=includes,
        exclude_unset=exclude_unset,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
    )

    expansion_depth = 0
    while expansions and expansion_depth < maximum_expansion_depth:
        expansions = await render_expansions(
            rendered_model,
            expansions,
            raise_error_on_expansion_not_found=raise_error_on_expansion_not_found,
            expansion_context=expansion_context,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        expansion_depth += 1

    return rendered_model


async def render_expansions(
    rendered_content: dict,
    expansions: Set[ExpansionInstruction],
    raise_error_on_expansion_not_found: bool,
    expansion_context: Any = None,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
) -> Set[ExpansionInstruction]:
    new_expansions = set()

    for expansion in expansions:
        expansion.future = expansion.expansion_definition.expand(  # type: ignore
            source_model=expansion.source_model, context=expansion_context
        )

    await gather(*[e.future for e in expansions if e.future])

    for expansion in expansions:
        expanded_value = None
        if expansion.future:
            expanded_value = await expansion.future

        if expanded_value is None:
            if raise_error_on_expansion_not_found:
                raise Exception(
                    f"Expansion `{'.'.join([str(p) for p in expansion.path])}` not found"
                )
            continue

        includes, sub_expansions = fieldset_to_includes(
            expansion.fieldsets, expanded_value
        )

        if expansion.expansion_definition.merge_fields_upwards and not isinstance(
            expanded_value, (BaseModel, dict)
        ):
            raise ValueError(
                f"Expansion `{expansion.expansion_name}` on "
                f"`{str(expansion.source_model.__class__)}` "
                "is defined with `merge_fields_upwards=True`, but the expansion did not return "
                "a pydantic.BaseModel or a dict.  Found `{type(expanded_value)}`."
            )

        # serialize new expanded values
        rendered_value = nested_structure_model_dump(
            value=expanded_value,
            includes=includes,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        if (
            expansion.expansion_definition.merge_fields_upwards
            and len(expansion.path) > 0
        ):
            path_put(rendered_content, copy(expansion.path)[:-1], rendered_value)
        else:
            path_put(rendered_content, copy(expansion.path), rendered_value)

        # Note that new expansions are not executed here nor are they done recursively here
        # we need to aggregate them up a level so that there is a chance to coalesce
        # calls to the same data loader.
        if sub_expansions:
            for sub_expansion in sub_expansions:
                sub_expansion.path = expansion.path + sub_expansion.path
            new_expansions.update(sub_expansions)

    return new_expansions


def nested_structure_model_dump(
    value: Any,
    includes: dict,
    exclude_unset: bool,
    exclude_defaults: bool,
    exclude_none: bool,
) -> Any:
    if isinstance(value, BaseModel):
        return value.model_dump(
            by_alias=False,
            include=includes,
            exclude=None,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    if isinstance(value, (list, set, tuple)):
        return [
            nested_structure_model_dump(
                v, includes[idx], exclude_unset, exclude_defaults, exclude_none
            )
            for idx, v in enumerate(value)
        ]

    if isinstance(value, (dict)):
        return {
            k: nested_structure_model_dump(
                v,
                includes.get(k) or {"__all__": {}},
                exclude_unset,
                exclude_defaults,
                exclude_none,
            )
            for k, v in value.items()
        }

    return value



import importlib.metadata

from .models import FieldsetConfig, ModelExpansion
from .render import render_fieldset_model
from .schema import FieldsetGenerateJsonSchema

try:
    __version__ = importlib.metadata.version("pydantic_enhanced_serializer")
except importlib.metadata.PackageNotFoundError:
    __version__ = "dev"



from collections import defaultdict
from inspect import isclass
from typing import Any, Dict, List, Optional, Set, Tuple, Union, get_origin

from pydantic import BaseModel

from .models import ExpansionBase, ExpansionInstruction


def fieldset_to_includes(
    fields_request: Union[str, Set[str], List[str]],
    model_data: Any,
    path: Optional[List[Union[str, int]]] = None,
    expansion_context: Any = None,
) -> Tuple[dict, Set[ExpansionInstruction]]:
    """
    Recursively descend a fieldsets list along with a pydantic model and produce:

    1) An `include` specification matching fieldsets that pydantic.dict will understand
    2) A list of requested expansions

    :parameters:
        - fields_request: Set of strings, each string is a field or fieldset name.  For nested
                     objects, names should be dot seperated: `top.submodel.subfield`.

        - model: A pydantic model INSTANCE, the top of the object tree that `fields_request` is
                 inteded to match

    :returns:
        - dict in pydantic `include` format meant for passing to model.dict(include=)
        - list of expansion actions needed

    :pydantic model.Config extensions:
        - :`fieldsets`: A dict of names (used as values in `fieldsets`) to one of
            - A list of strings.  These are field names as defined in the model.
            - An Expansion object from .expansions

        - `fieldsets["default"]`.  A set of fields to return when no specific request
           is made in `fieldsets` parameter.

        - If no `fieldsets` config value is present, then ALL fields will be returned
          (ie: normal pydantic.dict behavior).
    """
    includes: Dict[Union[str, int], Any] = defaultdict(dict)
    current_includes_ptr = includes
    expansions: Set[ExpansionInstruction] = set()
    expansion_fieldsets: Dict[str, Set[str]] = defaultdict(set)

    subfields: Set[str]

    if model_data is None:
        return {}, set()

    if isinstance(fields_request, str):
        fields_request = set(fields_request.split(","))

    if isinstance(fields_request, list):
        fields_request = set(fields_request)

    model = model_data
    if isinstance(model, list):
        for idx, submodel in enumerate(model):
            sub_includes, sub_expansions = fieldset_to_includes(
                fields_request, submodel, path + [idx] if path else [idx]
            )

            # If nothing under submodels is a base model, we need to add an
            # extra "__all__" for pydantic to pick up all the non-model data
            current_includes_ptr[idx].update(sub_includes or {"__all__": {}})
            expansions.update(sub_expansions)

        return {k: v for k, v in includes.items() if v is not None}, expansions

    if isinstance(model, dict):
        subfields = set([f.split(".", 1)[-1] for f in fields_request])

        for key, value in model.items():
            sub_includes, sub_expansions = fieldset_to_includes(
                subfields, value, path + [key] if path else [key]
            )

            current_includes_ptr[key] = sub_includes
            expansions.update(sub_expansions)

        return {k: v for k, v in includes.items() if v is not None}, expansions

    if not isinstance(model, BaseModel):
        return {}, set()

    fieldsets: Optional[dict] = getattr(model, "fieldset_config", {}).get(
        "fieldsets", None
    )
    default_fieldset: Optional[Set] = None
    if fieldsets and fieldsets.get("default"):
        default_fieldset = (
            set(fieldsets["default"])
            if isinstance(fieldsets["default"], list)
            else set([fieldsets["default"]])
        )

    if fieldsets is None or (default_fieldset and "*" in default_fieldset):
        # no fieldsets set or * in default, enable ALL fields
        fields_request.update([name for name in model.model_fields.keys()])

        # and add in all expansions
        if fieldsets:
            fields_request.update(
                [
                    name
                    for name in fieldsets.keys()
                    if isinstance(fieldsets[name], ExpansionBase)
                ]
            )

    elif default_fieldset:
        fields_request.update(default_fieldset)

    if path is None:
        path = []

    for fieldset in fields_request:
        if not fieldset:
            continue

        field: str

        try:
            field, subfield = fieldset.split(".", 1)
            subfields = set([subfield])
        except ValueError:
            field = fieldset
            subfields = set([])

        field_obj = model.model_fields.get(field)

        if field_obj:
            if (
                (origin := get_origin(field_obj.annotation))
                and isclass(origin)
                and issubclass(origin, (list, set, tuple))
            ):
                # Field value is a list of models
                if field not in current_includes_ptr:
                    current_includes_ptr[field] = defaultdict(dict)

                # while this could be done abstractly on the model class
                # and using __all__, we need to examine each item for its
                # own expansions
                for idx, item in enumerate(getattr(model, field) or []):
                    sub_includes, sub_expansions = fieldset_to_includes(
                        subfields, item, path + [field, idx]
                    )

                    current_includes_ptr[field][idx].update(sub_includes)
                    expansions.update(sub_expansions)

            elif (
                (origin := get_origin(field_obj.annotation))
                and isclass(origin)
                and issubclass(origin, (dict,))
            ):
                # Field is a dict, values may or may not contain models
                # or nested dicts/lists of models
                if field not in current_includes_ptr:
                    current_includes_ptr[field] = defaultdict(dict)

                for key, value in (getattr(model, field) or {}).items():
                    sub_includes, sub_expansions = fieldset_to_includes(
                        subfields, value, path + [field, key]
                    )

                    current_includes_ptr[field][key].update(sub_includes)
                    expansions.update(sub_expansions)

            else:
                # Field is a single model
                if field not in current_includes_ptr:
                    current_includes_ptr[field] = defaultdict(dict)

                sub_includes, sub_expansions = fieldset_to_includes(
                    subfields, getattr(model, field), path + [field]
                )

                current_includes_ptr[field].update(sub_includes)
                expansions.update(sub_expansions)

        elif (
            fieldsets
            and (expansion := fieldsets.get(field))
            and isinstance(expansion, ExpansionBase)
        ):
            if isinstance(model_data, list):
                # We need to create an expansion per list item
                for idx, source_model in enumerate(model_data):
                    expansions.add(
                        ExpansionInstruction(
                            expansion_definition=expansion,
                            expansion_name=field,
                            path=path + [idx, field],
                            fieldsets=list(subfields),
                            source_model=source_model,
                        )
                    )
            else:
                # as there may be multiple expansion fieldsets in the request, we need
                # to accumulate them first and then handle them later once all fieldset
                # requests have been seen
                expansion_fieldsets[field].update(subfields)

        elif fieldsets and (named_fieldset := fieldsets.get(field)):
            # Fieldset collection by name

            # reference to field of same name that does not exist leads to infinite recursion
            named_fieldset = [
                fieldset_field
                for fieldset_field in named_fieldset
                if (
                    model.model_fields.get(fieldset_field)
                    or (
                        fieldsets
                        and fieldsets.get(fieldset_field)
                        and field != fieldset_field
                    )
                )
            ]
            sub_includes, sub_expansions = fieldset_to_includes(
                named_fieldset, model, path
            )
            current_includes_ptr.update(sub_includes)
            expansions.update(sub_expansions)

    if expansion_fieldsets and fieldsets:
        for expansion, expansion_fields in expansion_fieldsets.items():
            expansion_definition = fieldsets.get(expansion)
            if not expansion_definition:
                continue

            expansions.add(
                ExpansionInstruction(
                    expansion_definition=expansion_definition,
                    expansion_name=expansion,
                    path=path + [expansion],
                    fieldsets=list(expansion_fields),
                    source_model=model,
                )
            )

    return {k: v for k, v in includes.items() if v is not None}, expansions



from __future__ import annotations

import abc
from asyncio import get_event_loop, isfuture
from typing import Any, Awaitable, Dict, List, Optional, TypedDict, Union

from pydantic import BaseModel, ConfigDict, Field


class ExpansionBase(BaseModel, abc.ABC):
    response_model: Optional[Any] = Field(
        default=None,
        description=(
            "Response model the expanded object will be cast to.  This is "
            "mainly used as a type hint by the schema generator"
        ),
    )
    merge_fields_upwards: Optional[bool] = Field(
        description=(
            "If true, the expander will attempt to merge items inside the "
            "object returned by `expand` into the parent object.  If false, "
            "then the name of expansion will become a key in the parent "
            "object with a value of the object returned by `expand`."
        ),
        default=False,
    )

    @abc.abstractmethod
    def expand(self, source_model: BaseModel, context: Any) -> Awaitable:
        """
        Seek out and expand based on a model.

        parameters:
            source_model:  A BaseModel instance from which to pick attributes
                           that drive the expansion.

            context:       The object passed in to render_fieldset_model.expansion_context

        Return:
            Any Awaitable that will resolve to:
                * A BaseModel instance
                * A scalar value
                * A list of the BaseModel or a list of scalar values
        """
        ...


class FieldsetConfig(TypedDict, total=False):
    fieldsets: Optional[Dict[str, List[str] | ExpansionBase]]


class ModelExpansion(ExpansionBase):
    """
    Expander will call:
        def expansion_method(self) -> response_model:

    Where self is the model on which the expansion was requested.
    """

    expansion_method_name: str

    def expand(self, source_model: BaseModel, context: Any) -> Awaitable:
        method = getattr(source_model, self.expansion_method_name, None)
        if not method:
            raise Exception(
                f"No such expansion method `{self.expansion_method_name}` found "
                f"on `{source_model.__class__}`"
            )

        value = method(context=context) if callable(method) else method
        if not isfuture(value):
            future = get_event_loop().create_future()
            future.set_result(value)
            return future

        return value


class ExpansionInstruction(BaseModel):
    expansion_definition: ExpansionBase
    expansion_name: str
    path: List[Union[str, int]]
    fieldsets: List[str]
    source_model: BaseModel

    # render time
    future: Optional[Awaitable] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __hash__(self) -> int:
        return sum([hash(str(p)) for p in self.path] + [hash(self.expansion_name)])



from inspect import isclass
from typing import Any, Dict, List, Optional, Set, Type, Union, get_args, get_origin

import pydantic
from packaging.version import Version
from packaging.version import parse as parse_version
from pydantic import BaseModel
from pydantic._internal._config import ConfigWrapper
from pydantic._internal._generate_schema import GenerateSchema
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from pydantic_core.core_schema import CoreSchema, ModelSchema

from .models import ExpansionBase

pydantic_version = parse_version(pydantic.__version__)
namespace_refactored_pydantic_version = Version("2.10")


class FieldsetGenerateJsonSchema(GenerateJsonSchema):
    def model_schema(self, schema: ModelSchema) -> JsonSchemaValue:
        json_schema = super().model_schema(schema)

        model = schema["cls"]

        # Not configured with fieldsets or expansions
        if not hasattr(model, "fieldset_config"):
            return json_schema

        fieldsets = model.fieldset_config.get("fieldsets")

        # for regular fields, set a description based on their fieldset configuration
        for field_name in model.model_fields.keys():
            if field_name not in json_schema.get("properties", {}):
                continue

            field_schema = json_schema["properties"][field_name]
            fieldset_names = [
                fieldset_name
                for fieldset_name in fieldsets.keys()
                if isinstance(fieldsets[fieldset_name], list)
                and (
                    field_name in fieldsets[fieldset_name]
                    or "*" in fieldsets[fieldset_name]
                )
            ]

            if not fieldset_names:
                # nothing is returned by default, must always ask for this field explicity
                field_schema["description"] = _concat_description(
                    field_schema.get("description"),
                    "Not returned by default.  Request this field by name.",
                )

            elif "default" not in fieldset_names and len(fieldset_names) > 0:
                field_schema["description"] = _concat_description(
                    field_schema.get("description"),
                    "Request by name or using fieldset(s): "
                    + ", ".join(
                        [f"`{f}`" for f in sorted(fieldset_names) if f != "default"]
                    )
                    + ".",
                )

        # detail expansions
        if pydantic_version < namespace_refactored_pydantic_version:
            generator = GenerateSchema(
                config_wrapper=ConfigWrapper(config={}), types_namespace=None  # type: ignore
            )
        else:
            generator = GenerateSchema(config_wrapper=ConfigWrapper(config={}))  # type: ignore

        for expansion_name, expansion in fieldsets.items():
            if not isinstance(expansion, ExpansionBase):
                continue

            if expansion.response_model is None:
                continue

            target_type = _get_target_type(expansion.response_model)

            # If this is a not before seen model class, it needs to be registered
            # before we can $ref it
            model_schema: Optional[CoreSchema] = None
            model_name = None
            sub_json_schema: Dict[str, Any] = {}

            if isclass(target_type) and issubclass(target_type, BaseModel):
                model_schema = generator._model_schema(target_type)
                defs_ref = self.get_defs_ref((model_schema["schema_ref"], self.mode))
                sub_json_schema = {"$ref": self.ref_template.format(model=defs_ref)}
                model_name = target_type.__pydantic_core_schema__.get("config", {}).get(
                    "title"
                )

                if _is_list(expansion.response_model):
                    sub_json_schema = {"type": "array", "items": sub_json_schema}

                elif _is_optional(expansion.response_model):
                    sub_json_schema = {"anyOf": [sub_json_schema, {"type": "null"}]}

                if defs_ref not in self.definitions:
                    # guard against recursion on the same object
                    self.definitions[defs_ref] = {}
                    self.generate_inner(
                        target_type.__pydantic_core_schema__  # type: ignore
                    )

            else:
                core_schema = generator.match_type(expansion.response_model)
                sub_json_schema = self.generate_inner(core_schema)

            json_schema["properties"][expansion_name] = {
                "title": (model_name or expansion_name).replace("_", " "),
                "description": f"Request by name or using fieldset(s): `{expansion_name}`.",
                **sub_json_schema,
            }

        return json_schema


def _concat_description(description: Optional[str], additional: str) -> str:
    if description is None:
        return additional

    if description.endswith("."):
        return " ".join([description, additional])
    else:
        return ". ".join([description, additional])


def _get_target_type(value: Any) -> Any:
    """Find the underlying contained type and its level of "structured" (dict/list) nesting"""
    if _is_optional(value):
        value = _get_optional_type(value)

    if not get_origin(value) or not isclass(get_origin(value)):
        return value

    if _is_list(value) and (list_args := get_args(value)):
        return _get_target_type(list_args[0])

    if (
        issubclass(get_origin(value), (dict, Dict))
        and (dict_args := get_args(value))
        and len(dict_args) == 2
    ):
        return _get_target_type(dict_args[1])

    return value


def _is_optional(type_: Any) -> bool:
    return (
        get_origin(type_) is Union
        and len(get_args(type_)) == 2
        and type(None) in get_args(type_)
    )


def _get_optional_type(type_: Any) -> Optional[Type]:
    for arg in get_args(type_):
        if arg is not None:
            return arg

    return None


def _is_list(type_: Any) -> bool:
    return bool(
        (origin := get_origin(type_))
        and isclass(origin)
        and issubclass(get_origin(type_), (list, List, set, Set))
    )


def model_has_fieldsets_defined(model: Any) -> bool:
    if _is_optional(model):
        model = _get_optional_type(model)

    if isclass(model) and issubclass(model, BaseModel):
        if getattr(model, "fieldset_config", None):
            return True

        else:
            return any(
                [
                    model_has_fieldsets_defined(field.annotation)
                    for field in model.model_fields.values()
                ]
            )

    return False



from typing import ClassVar, List, Optional

from pydantic import BaseModel

from pydantic_enhanced_serializer import (
    FieldsetConfig,
    FieldsetGenerateJsonSchema,
    ModelExpansion,
)


def test_fields_in_fieldset() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "extra": ["field1", "field2"],
                "extra2": ["field2", "field3"],
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert (
        schema["properties"]["field1"]["description"]
        == "Request by name or using fieldset(s): `extra`."
    )
    assert (
        schema["properties"]["field2"]["description"]
        == "Request by name or using fieldset(s): `extra`, `extra2`."
    )
    assert (
        schema["properties"]["field3"]["description"]
        == "Request by name or using fieldset(s): `extra2`."
    )


def test_fields_in_default() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field1", "field2"],
                "extra": ["field2", "field3"],
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "description" not in schema["properties"]["field1"]
    assert "description" not in schema["properties"]["field2"]
    assert (
        schema["properties"]["field3"]["description"]
        == "Request by name or using fieldset(s): `extra`."
    )


def test_fields_star_default() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["*"],
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "description" not in schema["properties"]["field1"]
    assert "description" not in schema["properties"]["field2"]
    assert "description" not in schema["properties"]["field3"]


def test_fields_named_default() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str
        field3: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"default": ["field1", "field2", "field3"]}
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "description" not in schema["properties"]["field1"]
    assert "description" not in schema["properties"]["field2"]
    assert "description" not in schema["properties"]["field3"]


def test_sub_object() -> None:
    class SubThing(BaseModel):
        sfield1: str
        sfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "f1": ["sfield1"],
                "f2": ["sfield1", "sfield2"],
            }
        )

    class Thing(BaseModel):
        field1: str
        field2: SubThing

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"default": ["field1", "field2"]}
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "description" not in schema["properties"]["field1"]
    assert "description" not in schema["properties"]["field2"]

    assert "SubThing" in schema["$defs"]
    assert (
        schema["$defs"]["SubThing"]["properties"]["sfield1"]["description"]
        == "Request by name or using fieldset(s): `f1`, `f2`."
    )
    assert (
        schema["$defs"]["SubThing"]["properties"]["sfield2"]["description"]
        == "Request by name or using fieldset(s): `f2`."
    )


def test_sub_object_list() -> None:
    class SubThing(BaseModel):
        sfield1: str
        sfield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "f1": ["sfield1"],
                "f2": ["sfield1", "sfield2"],
            }
        )

    class Thing(BaseModel):
        field1: str
        field2: List[SubThing]

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={"default": ["field1", "field2"]}
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "description" not in schema["properties"]["field1"]
    assert "description" not in schema["properties"]["field2"]

    assert "SubThing" in schema["$defs"]
    assert (
        schema["$defs"]["SubThing"]["properties"]["sfield1"]["description"]
        == "Request by name or using fieldset(s): `f1`, `f2`."
    )
    assert (
        schema["$defs"]["SubThing"]["properties"]["sfield2"]["description"]
        == "Request by name or using fieldset(s): `f2`."
    )


def test_expansion_model() -> None:
    class ExpandedThing(BaseModel):
        efield1: str
        efield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "f1": ["efield1", "efield2"],
                "f2": ["efield2"],
            }
        )

    from pydantic import Field

    class Thing(BaseModel):
        field1: str
        boo: ExpandedThing = Field(description="yo!")

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "expando": ModelExpansion(
                    expansion_method_name="foo",
                    response_model=ExpandedThing,
                )
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]
    assert schema["$defs"]
    assert "ExpandedThing" in schema["$defs"]
    assert "properties" in schema["$defs"]["ExpandedThing"]

    assert "expando" in schema["properties"]
    assert schema["properties"]["expando"] == {
        "title": "ExpandedThing",
        "description": "Request by name or using fieldset(s): `expando`.",
        "$ref": "#/$defs/ExpandedThing",
    }


def test_expansion_model_list() -> None:
    class ExpandedThing(BaseModel):
        efield1: str
        efield2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "f1": ["efield1", "efield2"],
                "f2": ["efield2"],
            }
        )

    class Thing(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "expando": ModelExpansion(
                    expansion_method_name="foo",
                    response_model=List[ExpandedThing],
                )
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "expando" in schema["properties"]
    assert schema["properties"]["expando"] == {
        "title": "ExpandedThing",
        "description": "Request by name or using fieldset(s): `expando`.",
        "type": "array",
        "items": {
            "$ref": "#/$defs/ExpandedThing",
        },
    }


def test_expansion_scalar() -> None:
    class Thing(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "expando": ModelExpansion(
                    expansion_method_name="foo",
                    response_model=int,
                )
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "expando" in schema["properties"]
    assert schema["properties"]["expando"] == {
        "title": "expando",
        "description": "Request by name or using fieldset(s): `expando`.",
        "type": "integer",
    }


def test_expansion_scalar_list() -> None:
    class Thing(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "expando": ModelExpansion(
                    expansion_method_name="foo",
                    response_model=List[int],
                )
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)
    assert schema
    assert schema["properties"]

    assert "expando" in schema["properties"]
    assert schema["properties"]["expando"] == {
        "title": "expando",
        "description": "Request by name or using fieldset(s): `expando`.",
        "type": "array",
        "items": {
            "type": "integer",
        },
    }


def test_optional_expansion_response_model() -> None:
    class Expanded(BaseModel):
        efield1: str

    class Thing(BaseModel):
        field1: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "expando": ModelExpansion(
                    expansion_method_name="foo", response_model=Optional[Expanded]
                )
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)

    assert schema
    assert schema["properties"]

    assert "expando" in schema["properties"]
    assert schema["properties"]["expando"] == {
        "title": "Expanded",
        "description": "Request by name or using fieldset(s): `expando`.",
        "anyOf": [{"$ref": "#/$defs/Expanded"}, {"type": "null"}],
    }


def test_unfieldseted_field_description() -> None:
    class Thing(BaseModel):
        field1: str
        field2: str

        fieldset_config: ClassVar = FieldsetConfig(
            fieldsets={
                "default": ["field1"],
            }
        )

    schema = Thing.model_json_schema(schema_generator=FieldsetGenerateJsonSchema)

    assert schema
    assert schema["properties"]

    assert "field2" in schema["properties"]
    assert schema["properties"]["field2"] == {
        "title": "Field2",
        "description": "Not returned by default.  Request this field by name.",
        "type": "string",
    }
